#!/usr/bin/env python

import math
import os

from PyQt5 import QtCore, QtGui, QtWidgets

import enums
import interfaces

location = os.path.dirname(os.path.realpath(__file__))
RES_IMAGES = location+'/resources/images/'
ICONS = {enums.DeviceType.datapath_access: 'datapath.png',
	 enums.DeviceType.datapath_distribution: 'datapath_distribution.png',
	 enums.DeviceType.datapath_core: 'datapath_core.png',
	 enums.DeviceType.node: 'node.png'}

class Link(QtWidgets.QGraphicsLineItem):
    def __init__(self, source, destination):
        super().__init__()
        self.source = source
        self.destination = destination
        self.stats = None

        self.setZValue(10)
        pen = QtGui.QPen(QtGui.QColor('#aaa'))
        self.setPen(pen)
        self.draw()

    def draw(self):
        r1 = self.source.mapRectToScene(self.source.boundingRect())
        r2 = self.destination.mapRectToScene(self.destination.boundingRect())
        
        c1 = r1.center()
        c2 = r2.center()

        x1 = c1.x()
        y1 = c1.y()
        x2 = c2.x()
        y2 = c2.y()

        self.setLine(x1, y1, x2, y2)
        if self.stats != None:
            self.position_stats()

    def attach_stats(self, stats):
        self.stats = stats
        self.position_stats()

    def position_stats(self):
        if isinstance(self.destination, Node):
            self.position_node_stats(self.destination)
        else:
            self.position_datapath_stats()

    def position_node_stats(self, node):
        node_pos = node.pos()
        x = node_pos.x()
        y = node_pos.y()

        stats_rect = self.stats.boundingRect()
        h = stats_rect.height()

        self.stats.setPos(x+10, y-h)
        self.stats.show()

    def position_datapath_stats(self):
        stats_rect = self.stats.boundingRect()
        w = stats_rect.width()/2
        h = stats_rect.height()/2

        line_center = self.boundingRect().center()
        x = line_center.x()
        y = line_center.y()

        self.stats.setPos(x-w, y-h)
        self.stats.show()        
        
    def update_stats(self, rx, tx):
        self.stats.set_values(rx, tx)

class TextLine(QtWidgets.QGraphicsSimpleTextItem):
    def __init__(self, text, size=16):
        super().__init__(text)
        self.setFont(QtGui.QFont('Ubuntu', size))

class Stats(QtWidgets.QGraphicsItemGroup):
    def __init__(self, source, destination):
        super().__init__()

        #self.source = source.uid.text()
        self.destination = destination

        self.setZValue(15)
        self.draw()

    def draw(self):
        self.dst_text = TextLine(self.destination.uid.text()+":", 14)
        self.dst_text.setPos(10, 10)
        self.addToGroup(self.dst_text)
        h = self.dst_text.boundingRect().height()

        self.rx = TextLine("RX: Measuring...", 14)
        self.rx.setPos(10, 12 + h)
        self.addToGroup(self.rx)
        self.tx = TextLine("TX: Measuring...", 14)
        self.tx.setPos(10, 12 + 2*h)
        self.addToGroup(self.tx)
        
        p = 7
        padding = [-p, -p, p, p]
        new_rect = self.boundingRect().adjusted(*padding)
        self.frame = QtWidgets.QGraphicsRectItem(new_rect)
        self.frame.setBrush(QtGui.QBrush(QtGui.QColor('#fcfcfc')))
        pen = QtGui.QPen(QtGui.QColor('#e0e0e0'))
        pen.setWidth(1)
        self.frame.setPen(pen)
        self.frame.setZValue(-1)
        self.addToGroup(self.frame)
        
    def set_values(self, rx, tx):
        rx = rx/1000
        tx = tx/1000

        self.rx.setText("RX: "+'%.3f'%round(rx, 3)+" kbps")
        self.tx.setText("TX: "+'%.3f'%round(tx, 3)+" kbps")

        self.dst_text.setText(self.destination.uid.text()+":")

        self.removeFromGroup(self.frame)
        new_rect = self.boundingRect().adjusted(-10, -10, 10, 10)
        self.frame.setRect(new_rect)
        self.addToGroup(self.frame)


class Element(QtWidgets.QGraphicsItemGroup):
    def __init__(self):
        super().__init__()
        self.selected = False

        self.setCacheMode(QtWidgets.QGraphicsItem.ItemCoordinateCache)
        self.links = set()
    
    def set_icon(self, name):
        self.icon = QtWidgets.QGraphicsPixmapItem(
            QtGui.QPixmap(RES_IMAGES+ICONS[name]).scaled(70, 70,
            transformMode=QtCore.Qt.SmoothTransformation))
        self.addToGroup(self.icon)

    def set_ip(self, ip):
        self.ip = TextLine(ip)
        h = self.icon.boundingRect().height()
        y = (h/2 - self.ip.boundingRect().height()) / 2 + h/2
        self.ip.setPos(90, y)
        self.addToGroup(self.ip)
    
    def set_uid(self, uid):
        self.uid = TextLine(uid)
        h = self.icon.boundingRect().height()
        y = (h/2 - self.uid.boundingRect().height()) / 2
        self.uid.setPos(90, y)
        self.addToGroup(self.uid)

    def create_frame(self):
        for item in self.childItems():
            item.moveBy(10, 10)
        new_rect = self.boundingRect().adjusted(0, 0, 20, 20)
        self.frame = QtWidgets.QGraphicsRectItem(new_rect)
        self.frame.setZValue(-1)

        pen = QtGui.QPen(QtGui.QColor('#bbbbbb'))
        pen.setWidth(2)
        self.frame.setPen(pen)

        self.addToGroup(self.frame)
        return self.frame

    def refresh_frame(self):
        self.removeFromGroup(self.frame)
        new_rect = self.boundingRect().adjusted(-10, -10, 10, 10)
        self.frame.setRect(new_rect)
        self.addToGroup(self.frame)

        for link in self.links:
            link.draw()

    def select(self):
        self.selected = True
        pen = QtGui.QPen(QtGui.QColor('orange'))
        pen.setWidth(3)
        self.frame.setPen(pen)

    def deselect(self):
        self.selected = False
        pen = QtGui.QPen(QtGui.QColor('#bbbbbb'))
        pen.setWidth(1)
        self.frame.setPen(pen)

    def rename(self, name):
        self.uid.setText(name)
        self.refresh_frame()


class Datapath(Element):
    def __init__(self, ip, dpid, level):
        super().__init__()

        self.level = level

        self.set_icon(self.level)
        self.set_uid(dpid)
        self.set_ip(ip)
        frame = self.create_frame()
        frame.setBrush(QtGui.QBrush(QtGui.QColor('#ffff88')))
        self.setZValue(30)

        
    def set_level(self, level):
        self.level = level

        icon = QtWidgets.QGraphicsPixmapItem(
            QtGui.QPixmap(RES_IMAGES+ICONS[level]).scaled(70, 70,
            transformMode=QtCore.Qt.SmoothTransformation))

        icon_pos = self.icon.pos()
        group_pos = self.pos()
        x = group_pos.x() + icon_pos.x()
        y = group_pos.y() + icon_pos.y()
        icon.setPos(x, y)

        self.removeFromGroup(self.icon)
        self.addToGroup(icon)
        self.icon = icon


class Node(Element):
    def __init__(self, ip, mac, parent_port):
        super().__init__()

        self.set_icon(enums.DeviceType.node)
        self.set_port(str(parent_port))
        self.set_uid(mac)
        self.set_ip(ip)
        frame = self.create_frame()
        frame.setBrush(QtGui.QBrush(QtGui.QColor('#ffffd4')))
        self.setZValue(20)

    def set_port(self, port):
        self.port = TextLine(port)
        if len(port) == 1:
            self.port.setPos(32, 14)
        else:
            self.port.setPos(21, 14)
        self.addToGroup(self.port)
