#!/usr/bin/env python

from enum import Enum
import collections

from PyQt5 import QtCore, QtGui, QtWidgets

import enums
import gui.canvas

class Mouse(Enum):
    drag_canvas = 0
    drag_item = 1


class View(QtWidgets.QGraphicsView):
    # object = gui.canvas.Element -> Datapath/Node
    elementMoved = QtCore.pyqtSignal(object, QtCore.QPointF) # ok
    elementDoubleClicked = QtCore.pyqtSignal(object)         # ok
    elementClicked = QtCore.pyqtSignal(object)               # ok

    createTree = QtCore.pyqtSignal()
    createGraph = QtCore.pyqtSignal()
    datapathLevelChanged = QtCore.pyqtSignal(object, object, object)

    def __init__(self):
        super().__init__()

        self.scale(0.6, 0.6)
        self.mouse_action = None
        self.network_levels = False
        self.path_definition = False

        self.setTransformationAnchor(QtWidgets.QGraphicsView.NoAnchor)
        self.setRenderHints(QtGui.QPainter.Antialiasing)
        self.setViewportUpdateMode(QtWidgets.QGraphicsView.SmartViewportUpdate)
        self.setCacheMode(QtWidgets.QGraphicsView.CacheBackground)

    def mousePressEvent(self, event):
        self.click_pos = event.pos()
        self.clicked_item = None
        self.clicked_element = None
        self.mouse_action = None

        self.last_pos = self.mapToScene(event.pos())
        if event.button() == QtCore.Qt.LeftButton:
            self.clicked_item = self.itemAt(event.pos())
            if self.clicked_item is not None:
                self.clicked_element = self.clicked_item.group()

            if isinstance(self.clicked_element, gui.canvas.Element):
                self.mouse_action = Mouse.drag_item
            else:
                self.mouse_action = Mouse.drag_canvas
            QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.OpenHandCursor)

    def mouseReleaseEvent(self, event):
        if self.mouse_action is Mouse.drag_item:
            if self.path_definition and self.click_pos == event.pos():
                self.elementClicked.emit(self.clicked_element)
            else:
                self.elementMoved.emit(self.clicked_element, self.last_pos)

        QtWidgets.QApplication.restoreOverrideCursor()

    def mouseMoveEvent(self, event):
        view_screen_start = self.mapToGlobal(QtCore.QPoint(0,0))
        view_screen_end = self.mapToGlobal(QtCore.QPoint(self.width(),
            self.height()))
        view_end = QtCore.QPoint(self.width(), self.height())

        pos = self.mapToScene(event.pos())
        x = pos.x() - self.last_pos.x()
        y = pos.y() - self.last_pos.y()

        if not (view_screen_start.x() <= event.screenPos().x() <=
                view_screen_end.x() and view_screen_start.y() <=
                event.screenPos().y() <= view_screen_end.y()):
            return None

        if self.mouse_action is Mouse.drag_item:
            if self.clicked_element.isActive():
                if (QtWidgets.QApplication.overrideCursor().shape() !=
                        QtCore.Qt.ClosedHandCursor):
                    QtWidgets.QApplication.changeOverrideCursor(
                        QtCore.Qt.ClosedHandCursor)

                rect_start = self.mapFromScene(
                    self.clicked_element.mapToScene(
                    self.clicked_element.boundingRect().topLeft()))
                rect_end = self.mapFromScene(
                    self.clicked_element.mapToScene(
                    self.clicked_element.boundingRect().bottomRight()))

                self.clicked_element.moveBy(x, y)
                self.last_pos = self.mapToScene(event.pos())
                # zavolat move_element az v mouseRelease a ulozit finalni pozici
                #self.elementMoved.emit(self.clicked_element, self.last_pos)

                band = self.width() * 0.05
                x = 0
                y = 0
                
                if rect_start.x() < band:
                    x = 1
                elif (view_end.x() - rect_end.x()) < band:
                    x = -1
                if rect_start.y() < band:
                    y = 1
                elif (view_end.y() - rect_end.y()) < band:
                    y = -1

                if x or y:
                    self.translate(20*x, 20*y)

                for link in self.clicked_element.links:
                    link.draw()
            else:
                QtWidgets.QApplication.restoreOverrideCursor()
        elif self.mouse_action is Mouse.drag_canvas:
            self.translate(x, y)

    def mouseDoubleClickEvent(self, event):
        clicked_item = self.itemAt(event.pos())
        if clicked_item is not None:
            clicked_element = clicked_item.group()
            if isinstance(clicked_element, gui.canvas.Element):
                self.elementDoubleClicked.emit(clicked_element)

    def contextMenuEvent(self, event):
        clicked_item = self.itemAt(event.pos())
        clicked_element = (clicked_item.group() if clicked_item is not None
            else None)

        menu = QtWidgets.QMenu();
        if isinstance(clicked_element, gui.canvas.Element):
            if isinstance(clicked_element, gui.canvas.Node):
                return
            elif isinstance(clicked_element, gui.canvas.Datapath):
                group_level = QtWidgets.QActionGroup(menu)
                group_level.triggered.connect(
                    self.emit_datapath_level_changed)

                action_access = QtWidgets.QAction("Access", group_level)
                action_distribution = QtWidgets.QAction("Distribution",
                                                        group_level)
                action_core = QtWidgets.QAction("Core", group_level)

                actions = collections.OrderedDict()
                actions[enums.DeviceType.datapath_access] =  action_access
                actions[enums.DeviceType.datapath_distribution] = (
                    action_distribution)
                actions[enums.DeviceType.datapath_core] = action_core
                for level, action in actions.items():
                    action.setCheckable(True)
                    action.setData({'element': clicked_element,
                                    'old_level': clicked_element.level,
                                    'new_level': level})
                actions[clicked_element.level].setChecked(True)

                menu.addActions(actions.values())
        else:
            action_tree = QtWidgets.QAction("Generate tree", menu)
            if not self.network_levels:
                action_tree.setEnabled(False)
            menu.addAction(action_tree)
            action_tree.triggered.connect(self.emit_create_tree)

            action_random_graph = QtWidgets.QAction("Generate graph", menu)
            menu.addAction(action_random_graph)
            action_random_graph.triggered.connect(
						 self.emit_create_graph)

        menu.exec(event.globalPos())

    #def path_definition_state_changed(self, state):
    #    self.path_definition = state

    def network_levels_state_changed(self, state):
        self.network_levels = state

    def path_definition_state_changed(self, state):
        self.path_definition = state

    ### nahradit one-linery lambda funkcemi
    ### nebo tlacitka udelat jako self_ a pak se k signalu dostanu zvenku
    def emit_create_tree(self):
        self.createTree.emit()

    def emit_create_graph(self):
        self.createGraph.emit()

    def emit_datapath_level_changed(self, action):
        data = action.data()
        self.datapathLevelChanged.emit(data['element'],
            data['old_level'], data['new_level'])
