#!/usr/bin/env python

"""
Module 'base' serves as a base for communication modules for SDN controllers.

This module defines classes that are to ensure communication with an SDN
controller and store retrieved data. It is of an abstract nature and not meant
to be used itself. Instead, it should be used to create modules tailored
for particular controllers.
"""

# abc is a module for defining abstract base classes
import abc

class BaseNetwork(metaclass=abc.ABCMeta):
    """Defines methods for controller communication.

       This abstract class is to be subclassed by a controller-specific
       communication interface. It defines methods that have to be implemented,
       and gathers data structures for storing data retrieved from the
       controller.
    """

    def __init__(self):
        """Initialize instance variables."""
        self._datapaths = {}
        self._nodes = {}
        self._links = set()

    @abc.abstractmethod
    def connect(self, server, user, password):
        """Connect to a controller.

        This method has to be overriden.

        Parameters
        ----------
        server : string
            IP address of the server running the controller software.
        user : string
            Username to log in to the controller.
        password : string
            Password to log in to the controller.
        """
        pass

    @abc.abstractmethod
    def load(self):
        """Load network information into instance variables.

        This method has to be overriden.
        """
        pass
            
    @abc.abstractmethod
    def update(self, interval):
        """Update network information and return the differences.

        This method has to be overriden.

        Parameters
        ----------
        interval : int
            Interval in which network updates are performed.

        Returns
        -------
        Update
            A summary of changes in the network - devices to add or remove,
            links to connect or disconnect.
        """
        pass
     
    @abc.abstractmethod
    def add_flow(self, datapath, src_ip, dst_ip, in_port, out_port,
                 priority, idle_timeout, hard_timeout, **kwargs):
        """Push a flow entry of the given parameters into the given datapath.

        This method has to be overriden.

        Parameters
        ----------
        datapath : Datapath
            Datapath to install the new flow entry.
        src_ip : string
            Source IPv4 address to match.
        dst_ip : string
            Destination IPv4 address to match.
        in_port : int
            Packet ingress port to match.
        out_port : int
            Port to send the packet out.
        priority : int
            Priority (0-65535).
        idle_timeout : int
            Idle timeout (0-65535). 
        hard_timeout : int
            Hard timeout (0-65535).
        **kwargs
            Listed below.

        Keyword arguments
        -----------------
        protocol : enums.IpProtocol
            IP protocol.
        tcp_src : int
            TCP source port.
        tcp_dst : int
            TCP destination port.
        udp_src : int
            UDP source port.
        udp_dst : int
            UDP destination port.
        """
        pass

    @property
    def datapaths(self):
        """Get list of Datapath objects."""
        return list(self._datapaths.values())

    def datapath_by_dpid(self, dpid):
        """Return a Datapath object for the given DPID.

        Returns
        -------
        Datapath
            When a datapath with the given DPID exists in the network.
        None
            When a datapath with the given DPID does not exist in the network.
        """
        if dpid in self._datapaths:
            return self._datapaths[dpid]
        else:
            return None

    def node_by_mac(self, mac):
        """Return a Node object for the given MAC address.

        Returns
        -------
        Node
            When a node with the given MAC address exists in the network.
        None
            When a node with the given MAC address does not exist
            in the network.
        """
        if mac in self._nodes:
            return self._nodes[mac]
        else:
            return None

    @property
    def nodes(self):
        """Get list of Node objects."""
        return list(self._nodes.values())

    @property
    def links(self):
        """Get set of Link objects."""
        return self._links.copy()


class Datapath:
    """A representation of a networking device - a datapath."""
    def __init__(self, dpid, ip, of_version):
        """Initialize instance variables.

        Parameters
        ----------
        dpid : string
            Datapath DPID.
        ip : string
            Datapath IP address.
        of_version : string
            Negotiated version of the OpenFlow protocol.
        """
        self.dpid = dpid
        self.ip = ip
        self.of_version = of_version
        self.ports_devices = {}  # port: Datapath or Node
        self.ports_stats = {}    # port: PortStats

    def connect(self, port, device):
        """Connect the given device to the given port.

        Parameters
        ----------
        port : int
            Port number.
        device : Datapath or Node
            Device to connect.
        """
        self.ports_devices[port] = device
        self.ports_stats[port] = PortStats()

    def disconnect_port(self, port):
        """Disconnect the device connected to the given port.

        Parameters
        ----------
        port : int
            Port number.
        """
        if port in self.ports_devices:
            del self.ports_devices[port]
            del self.ports_stats[port]

    def disconnect_device(self, device):
        """Disconnect the given device from the datapath.

        Parameters
        ----------
        device : Datapath or Node
            Device to disconnect.
        """
        port = self.port_of_device(device)
        self.disconnect_port(port)

    def has_connected(self, device):
        """Return whether the given device is connected to the datapath.

        Parameters
        ----------
        device : Datapath or Node
            Device to inquire.

        Returns
        -------
        bool
            True when the device is connected to the datapath, False otherwise.
        """
        return device in self.ports_devices.values()

    def port_of_device(self, device):
        """Return the port to which the given device is connected.

        Parameters
        ----------
        device : Datapath or Node
            Device to inquire.

        Returns
        -------
        int
            Port number.
        None
            In case the device is not connected to the datapath.
        """
        for (port, port_device) in self.ports_devices.items():
            if port_device == device:
                return port
        return None

    def port_stats(self, port):
        """Return port statistics for the given port.

        Parameters
        ----------
        port : int
            Port number.

        Returns
        -------
        PortStats
            Port statistics.
        None
            In case the port is not alive.
        """
        if port in self.ports_stats:
            return self.ports_stats[port]
        else:
            return None
    
    def update_port_stats(self, port, interval, rx_bytes, tx_bytes):
        """Update port statistics for the given port.

        Parameters
        ----------
        port : int
            Port number.
        interval : int
            Interval between individual updates in seconds.
        rx_bytes : int
            Received bytes.
        tx_bytes : int
            Transmitted bytes.
        """
        ps = self.ports_stats[port]
        # In case the number of bytes is zero, update it and wait
        # for the next update with rate calculation
        if ps.rx_bytes != 0:
            ps.rx_rate = (rx_bytes - ps.rx_bytes) / interval * 8
        else:
            ps.rx_rate = 0
        if ps.tx_bytes != 0:
            ps.tx_rate = (tx_bytes - ps.tx_bytes) / interval * 8
        else:
            ps.tx_rate = 0
        ps.rx_bytes = rx_bytes
        ps.tx_bytes = tx_bytes

    def __hash__(self):
        return hash((self.dpid, self.ip, self.of_version))

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        else:
            return ((self.dpid, self.ip, self.of_version) ==
                   (other.dpid, other.ip, other.of_version))


class Node:
    """A representation of an end-node device."""
    def __init__(self, mac, ip, parent, parent_port):
        """Initialize instance variables.

        Parameters
        ----------
        mac : string
            Node MAC address.
        ip : string
            Node IP address.
        parent : Datapath
            Parent datapath of the Node.
        parent_port : int
            Parent datapath port to which the Node connects.
        """
        self.mac = mac
        self.ip = ip
        self.parent = parent
        self.parent_port = parent_port

    def __hash__(self):
        return hash((self.mac, self.ip, self.parent, self.parent_port))

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        else:
            return ((self.mac, self.ip, self.parent, self.parent_port) ==
                    (other.mac, other.ip, other.parent, other.parent_port))


class Link:
    """A representation of a connection between two devices."""
    def __init__(self, src, dst):
        """Initialize instance variables.

        Parameters
        ----------
        src : Datapath
            Source datapath.
        dst : Datapath or Node
            Destination device.
        """
        self.src = src
        self.dst = dst

    def __hash__(self):
        return hash(frozenset({self.src, self.dst}))

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        else:
            return {self.src, self.dst} == {other.src, other.dst}


class PortStats:
    """A data container to hold statistics for a datapath port."""
    def __init__(self):
        """Initialize instance variables."""
        self.rx_bytes = 0
        self.tx_bytes = 0
        self.rx_rate = 0  # bits per second
        self.tx_rate = 0  # bits per second


class Update:
    """A data container to hold changes in the network for a single update."""
    def __init__(self):
        """Initialize instance variables."""
        self.add = []
        self.remove = []
        self.connect = []
        self.disconnect = []
