#!/usr/bin/env python

"""
Module 'hpe' is an interface for communication with the HPE VAN SDN Controller.

This module is based on the 'base' module - it implements the BaseNetwork
abstract class and its abstract methods. It implements the functionality
to establish a connection with the controller, to retrieve required data,
and to push data back to the controller.
"""

import hpsdnclient as hp

import enums
import interfaces.base as base

FULL_NAME = "HPE VAN SDN Controller"

class Network(base.BaseNetwork):
    """Defines methods for controller communication."""

    def __init__(self):
        super().__init__()

    def connect(self, server, user, password):
        """Connect to a controller.

        This method has to be overriden.

        Parameters
        ----------
        server : string
            IP address of the server running the controller software.
        user : string
            Username to log in to the controller.
        password : string
            Password to log in to the controller.
        """
        # Obtain an authentication token
        auth = hp.XAuthToken(server=server, user=user, password=password)
        # Keep a reference to the API accessed with the token
        self._api = hp.Api(controller=server, auth=auth)

    def load(self):
        """Load network information into instance variables."""
        # Retrieve datapaths from the controller
        datapaths = self._api.get_datapaths()
        # Retrieve nodes from the controller
        nodes = self._api.get_nodes()
        # Retrieve links from the controller
        links = self._api.get_links()

        for dp in datapaths:
            # Save the datapath
            self._datapaths[dp.dpid] = base.Datapath(dp.dpid, dp.device_ip,
                                                     dp.negotiated_version)
        
        for n in nodes:
            # Save the node
            parent = self._datapaths[n.dpid]
            node = base.Node(n.mac, n.ip, parent, n.port)
            self._nodes[n.mac] = node
            # Connect the node to its parent datapath
            parent.connect(n.port, node)
            # Save the link
            link = base.Link(parent, node)
            self._links.add(link)

        for l in links:
            # Save the link
            link = base.Link(self._datapaths[l.src_dpid],
                             self._datapaths[l.dst_dpid])
            self._links.add(link)
            # Mutually connect the involved datapaths
            self._datapaths[l.src_dpid].connect(l.src_port,
                                                self._datapaths[l.dst_dpid])
            self._datapaths[l.dst_dpid].connect(l.dst_port,
                                                self._datapaths[l.src_dpid])

    def update(self, interval):
        """Update network information and return the differences.

        Parameters
        ----------
        interval : int
            Interval in which network updates are performed.

        Returns
        -------
        Update
            A summary of changes in the network - devices to add or remove,
            links to connect or disconnect.
        """
        update = base.Update()

        # Retrieve datapaths from the controller
        datapaths = self._api.get_datapaths()
        # Retrieve nodes from the controller
        nodes = self._api.get_nodes()

        # Prepare sets that will serve to compare the new network state to the
        # original state
        new_dpids = set()
        new_macs = set()
        new_links = set()
        
        # Iterate retrieved datapaths
        for dp in datapaths:
            new = base.Datapath(dp.dpid, dp.device_ip, dp.negotiated_version)
            new_dpids.add(dp.dpid)

            # Is this datapath (with the same DPID) already stored?
            # Does the new object differ from the original object?
            # Remove the original datapath if it differs from the new one.
            if dp.dpid in self._datapaths and self._datapaths[dp.dpid] != new:
                update.remove.append(self._datapaths[dp.dpid])
                del self._datapaths[dp.dpid]
 
            # Save the datapath in case it is not stored
            # Also applies if it was removed in the previous step
            if dp.dpid not in self._datapaths:
                update.add.append(new)
                self._datapaths[dp.dpid] = new

        # Remove datapaths that were in the original network state, but are not
        # in the new state
        leftover_dp = self._datapaths.keys() - new_dpids
        for i in leftover_dp:
            update.remove.append(self._datapaths[i])
            del self._datapaths[i]

        # Iterate retrieved nodes
        for n in nodes:
            new = base.Node(n.mac, n.ip, self._datapaths[n.dpid], n.port)
            new_macs.add(n.mac)

            # Is this node (with the same MAC address) already stored?
            # Does the new object differ from the original object?
            # Remove the original node if it differs from the new one.
            if n.mac in self._nodes and self._nodes[n.mac] != new:
                self._nodes[n.mac].parent.disconnect_device(self._nodes[n.mac])
                update.remove.append(self._nodes[n.mac])
                del self._nodes[n.mac]

            # Save and connect the node in case it is not stored
            # Also applies if it was removed in the previous step
            if n.mac not in self._nodes:
                self._nodes[n.mac] = new
                update.add.append(new)
                new.parent.connect(n.port, new)

            link = base.Link(new.parent, new)
            new_links.add(link)
            # Save the link in case it is not already stored
            if link not in self._links:
                self._links.add(link)
                update.connect.append(link)

        # Remove and disconnect nodes that were in the original network state,
        # but are not in the new state
        leftover_n = self._nodes.keys() - new_macs
        for i in leftover_n:
            self._nodes[i].parent.disconnect_device(self._nodes[i])
            update.remove.append(self._nodes[i])
            del self._nodes[i]

        # Iterate datapaths in the new state
        for dpid in self._datapaths: 
            # Retrieve links for this datapath from the controller
            links = self._api.get_links(dpid)
            # Prepare a set that will serve to compare the connected ports on
            # this datapath from the new network state to the ones from
            # the original state
            new_ports = set()
            for li in links:
                # Find out whether this datapath is in the link as source
                # or as destination
                if dpid == li.src_dpid:
                    this = li.src_dpid
                    this_port = li.src_port
                    other = li.dst_dpid
                    other_port = li.dst_port
                else:
                    this = li.dst_dpid
                    this_port = li.dst_port
                    other = li.src_dpid
                    other_port = li.src_port

                link = base.Link(self._datapaths[li.src_dpid],
                                 self._datapaths[li.dst_dpid])
                new_links.add(link)
                new_ports.add(this_port)

                # Connect the datapaths
                # In case the other datapath has been already connected and
                # its port has not changed, do not connect the datapaths again
                if (self._datapaths[this].port_of_device(
                        self._datapaths[other]) != this_port):
                    self._datapaths[li.src_dpid].connect(
                        li.src_port, self._datapaths[li.dst_dpid])
                    self._datapaths[li.dst_dpid].connect(
                        li.dst_port, self._datapaths[li.src_dpid])

                # Save the link in case it is not already stored
                if link not in self._links:
                    self._links.add(link)
                    update.connect.append(link)
            
            # Clear ports that were connected to a datapath in the old state,
            # but are not anymore in the new state
            leftover_p = self._datapaths[dpid].ports_devices.keys() - new_ports
            for p in leftover_p:
                if isinstance(self._datapaths[dpid].ports_devices[p],
                              base.Datapath):
                    self._datapaths[dpid].disconnect_port(p)
       
            # Retrieve port statistics from the controller 
            stats = self._api.get_port_stats(dpid)
            for s in stats:
                # Update port statistics for each active port.
                # The check is necessary because a local port for out-of-band
                # switch-controller communication would not be included in
                # the ports, but would have its statistics.
                if s.port_id in self._datapaths[dpid].ports_stats:
                    self._datapaths[dpid].update_port_stats(s.port_id,
                        interval, s.rx_bytes, s.tx_bytes)

        # Remove links that were in the original network state, but are not
        # in the new state
        leftover_li = self._links - new_links
        for i in leftover_li:
            update.disconnect.append(i)
        self._links = new_links

        return update

    def add_flow(self, datapath, src_ip, dst_ip, in_port, out_port,
                 priority, idle_timeout, hard_timeout, **kwargs):
        """Push a flow entry of the given parameters into the given datapath.

        Parameters
        ----------
        datapath : Datapath
            Datapath to install the new flow entry.
        src_ip : string
            Source IPv4 address to match.
        dst_ip : string
            Destination IPv4 address to match.
        in_port : int
            Packet ingress port to match.
        out_port : int
            Port to send the packet out.
        priority : int
            Priority (0-65535).
        idle_timeout : int
            Idle timeout (0-65535). 
        hard_timeout : int
            Hard timeout (0-65535).
        **kwargs
            Listed below.

        Keyword arguments
        -----------------
        protocol : enums.IpProtocol
            IP protocol.
        tcp_src : int
            TCP source port.
        tcp_dst : int
            TCP destination port.
        udp_src : int
            UDP source port.
        udp_dst : int
            UDP destination port.
        """
        # Prepare match fields of the flow entry
        match_args = {}
        match_args['ipv4_src'] = src_ip
        match_args['ipv4_dst'] = dst_ip
        match_args['eth_type'] = 'ipv4'
        match_args['in_port'] = in_port

        # Add fields related to IP protocol if given
        if kwargs['ip_protocol'] is enums.IpProtocol.tcp:
            match_args['ip_proto'] = 'tcp'
            if kwargs['tcp_src']:
                match_args['tcp_src'] = kwargs['tcp_src']
            if kwargs['tcp_dst']:
                match_args['tcp_dst'] = kwargs['tcp_dst']
        elif kwargs['ip_protocol'] is enums.IpProtocol.udp:
            match_args['ip_proto'] = 'udp'
            if kwargs['udp_src']:
                match_args['udp_src'] = kwargs['udp_src']
            if kwargs['udp_dst']:
                match_args['udp_dst'] = kwargs['udp_dst']

        # Convert match fields to the Match datatype of hp-sdn-client
        match = hp.datatypes.Match(**(match_args))
        # Create an action using the Action datatype of hp-sdn-client
        action = hp.datatypes.Action(output=out_port)

        # Prepare parameters of the flow entry
        flow_args = {}
        flow_args['priority'] = priority
        flow_args['idle_timeout'] = int(idle_timeout)
        flow_args['hard_timeout'] = int(hard_timeout)

        # Construct different messages for different versions of OpenFlow
        # using the Flow datatype of hp-sdn-client
        if datapath.of_version == "1.0.0":
            # Version 1.0.0 uses Actions
            flow = hp.datatypes.Flow(match=match, actions=action, **flow_args)
        else:
            # Version 1.3.0 uses Instructions with Actions inside
            instruction = hp.datatypes.Instruction(apply_actions=action)
            flow = hp.datatypes.Flow(match=match, instructions=[instruction],
                                     **flow_args)

        # Push the flow entry to the datapath
        self._api.add_flows(datapath.dpid, flow)
